kartsystem = {}
local karts = {}
local world = love.physics.newWorld(0, 0, true)
love.physics.setMeter(16)

local kartcount

local timer = 0

local phys = {}
local coins = {}

local canvas = {}

local tex = {}
local track = {}

local players = {}

local kartquad = {}
local particle = {}
local particles = {}
local splashes = {}
local raindrops = {}
local hazards = {}
local camera = {}
local doppler = {}

local sfx = {}

local rendersize = {}

local tracktype = {}
for i,j in pairs({
-- colors recognized
	asphault = {0.5, 0.5, 0.5},
	dirt =     {0.0, 1.0, 0.0},
	rumble =   {1.0, 0.0, 0.0},
	water =    {0.0, 0.0, 1.0},
	lava =     {1.0, 0.5, 1.0},
	bumper =   {1.0, 1.0, 0.0},
}) do
-- colors are converted to an integer via r*18+g*6+b*2
	tracktype[j[1]*18 + j[2]*6 + j[3]*2] = i
end


local function newRes(w,h,bg)
	rendersize = {w = w, h = h, bg = bg}
	rendersize.track = rendersize.h - rendersize.bg

	canvas.screen = {}
	for i = 1,playercount do
		canvas.screen[i] = lg.newCanvas(rendersize.w,rendersize.h)
	end
	canvas.trackdummy = lg.newCanvas(rendersize.w,rendersize.track)
	canvas.cloudsdummy = lg.newCanvas(rendersize.w,rendersize.bg)

	shader.clouds:send("mul", (rendersize.w/rendersize.h)*1.5)
end

function kartsystem.init(options)

	--fixedfps = options.fixed_fps
	playercount = options.players
	kartcount = options.players + options.ai

	track = {
		name = options.track,
		dir = "tracks/" .. options.track .. "/",
		bg = {},
		tex = {},
	}
	track.terrain = love.image.newImageData(track.dir .. "terrain.png")

	local trackconf = love.filesystem.load(track.dir .. "conf.lua")()
	for i,k in pairs(trackconf) do track[i] = k end

	for i,k in pairs(love.filesystem.getDirectoryItems(track.dir .. "tex")) do
		if k:sub(-4) == ".png" then
			track.tex[k:sub(1,-5)] = lg.newImage(track.dir .. "tex/" .. k)
		end
	end

	track.w = track.tex.main:getWidth()
	track.h = track.tex.main:getHeight()

	for i,k in pairs({"far", "sky", "lava"}) do
		if track.tex[k] then track.tex[k]:setWrap("repeat","repeat") end
	end

	for i,j in pairs(love.filesystem.getDirectoryItems(track.dir .. "/tex/bg")) do
		local name = tonumber(j:sub(1,-5))
		track.bg[name] = {}
		local b = track.bg[name]
		b.tex = lg.newImage(track.dir .. "tex/bg/" .. j)
		b.w = b.tex:getWidth()
		b.h = b.tex:getHeight()
	end


	tex = {
		ripple = lg.newImage("tex/ripple.png"),
		coin = lg.newImage("tex/coin.png"),
		hazard = {}
	}


	hazards = {}

	local points = love.filesystem.load(track.dir .. "points.lua")()
	hazards = points.hazards
	for i,j in pairs(love.filesystem.getDirectoryItems(track.dir .. "tex/hazards")) do
		tex.hazard[i] = lg.newImage(track.dir .. "tex/hazards/" .. j)
	end

	players = {
		{name = options.player1kart},
		{name = options.player2kart},
		{name = "bowser"},
		{name = "dkjr"},
		{name = "princess"},
		{name = "toad"},
		{name = "yoshi"},
		{name = "koopa"},
	}

	for i = 1,#players do
		players[i].tex = lg.newImage("karts/" .. players[i].name .. "/tex.png")
	end
	kartquad = {}
	for i =  0,21 do  kartquad[i] = lg.newQuad(i*32, 0, 32, 32, 32*22, 32) end



	particle = {}
	for i,k in pairs(love.filesystem.getDirectoryItems("tex/particles/")) do
		local name = k:sub(1,-5)
		particle[name] = {
			tex = lg.newImage("tex/particles/" .. k),
			quads = {}
		}
		particle[name].offset = particle[name].tex:getHeight() * 0.5 -- assumes square particles
		local w,h = particle[name].tex:getWidth(),particle[name].tex:getHeight()
		for i = 0,w/h - 1 do
			particle[name].quads[w/h - i] = lg.newQuad(i*h, 0, h, h, w, h)
		end
	end
	particles = {}

	splashes = {}


	local t = options.resolution
	if options.players == 2 then
		local ss = options.splitscreen
		if ss == "tall" then
			t.w = t.w * .5
		elseif ss == "wide" then
			--t.bgscale = t.bgscale * 0.5
			t.bg = t.bg*0.5
			t.fov = t.fov*2
			t.offset = 1 - (1 - t.offset)*0.5
			t.h = t.h * .5
		end
	end
	newRes(t.w,t.h,t.bg)

	camera = {
		offset = t.offset,
		fov = t.fov,
		bgscale = t.bgscale,
		kartscale = t.kartscale,
		splitscreen = options.splitscreen,
	}
	camera.zoom = (rendersize.w/4)*camera.offset / t.kartscale

	sfx = {}

	for p = 1,playercount do
		sfx[p] = {}
		for k = 1,kartcount do
			sfx[p][k] = {}
			local sfx = sfx[p][k]
			for j,f in pairs(love.filesystem.getDirectoryItems("sfx/loop/")) do
				local name = f:sub(1,-5)
				sfx[name] = love.audio.newSource("sfx/loop/" .. f, "static")
				sfx[name]:setLooping(true)
				sfx[name]:play()
				sfx[name]:seek(random(sfx[name]:getDuration()*1000, 1000)/1000)
				sfx[name]:setVolume(0)
			end
			for j,f in pairs(love.filesystem.getDirectoryItems("sfx/static/")) do
				local name = f:sub(1,-5)
				sfx[name] = love.audio.newSource("sfx/static/" .. f, "static")
			end
		end
	end

	shader.track:send("map", track.tex.main)
	shader.track:send("w", track.w)
	shader.track:send("h", track.h)
	shader.track:send("fogcol", track.fogcol)

	if track.tex.sky then shader.clouds:send("map", track.tex.sky) end

	shader.fog:send("fogcol", track.fogcol)

	shader.far:send("map", track.tex.far)
	shader.far:send("w", track.w)
	shader.far:send("h", track.h)
	shader.far:send("fogcol", track.fogcol)

	phys = points.physics
	coins = points.coins

	-- track physics
	track.shape = {}
	for i,k in ipairs(phys) do
		if #k == 4 then
			-- box2d doesn't like negative sizes
			if k[1] > k[3] then k[3],k[1] = k[1],k[3] end
			if k[2] > k[4] then k[2],k[4] = k[4],k[2] end
			local t = {
				body = love.physics.newBody(world, k[1] + (k[3]-k[1])*.5, k[2] + (k[4]-k[2])*.5, "static"),
				shape = love.physics.newRectangleShape(0, 0, k[3] - k[1], k[4] - k[2]),
			}
			t.fixture = love.physics.newFixture(t.body, t.shape)
			track.shape[#track.shape + 1] = t
		elseif #k == 2 then
			error("Lone vertex in track data, check physics editor")
		elseif #k > 6 then
			error("Polygons not implemented yet")
		end
	end

	track.ai = {}
	for i,k in ipairs(points.ai) do
		if #k == 6 then
			-- *we* don't like negative sizes
			if k[1] > k[3] then k[3],k[1] = k[1],k[3] end
			if k[2] > k[4] then k[2],k[4] = k[4],k[2] end
			track.ai[i] = k
		else
			error("AI Box #" .. i .. " doesn't have enough data, check map editor")
		end
	end

	-- hazard physics
	if love.filesystem.getInfo(track.dir .. "hazard.lua") then
		track.hazard = love.filesystem.load(track.dir .. "hazard.lua")()
		for i,k in ipairs(hazards) do
			local info = track.hazard[k[3]]
			if not info then error("No info in hazard.lua for hazard #" .. k[3]) end
			-- set the scale for each hazard, faster than looking it up each frame
			hazards[i][4] = info.size
			if info.physics.type == "circle" then
				local t = {
					body = love.physics.newBody(world, k[1], k[2], "static"),
					shape = love.physics.newCircleShape(info.physics.radius),
				}
				t.fixture = love.physics.newFixture(t.body, t.shape)
				track.shape[#track.shape + 1] = t
			end
		end
	end

	for p =1,playercount do
		doppler[p] = {}
		for k = 1,kartcount do
			-- last kart position relative to listener (player)
			doppler[p][k] = {x = 0, y = 0}
		end
	end

	canvas.track = lg.newCanvas(track.w, track.h)
	canvas.shadow = lg.newCanvas(track.w, track.h)
	canvas.tiremarks = lg.newCanvas(track.w, track.h)
	canvas.track:setWrap("clampzero","clampzero")

	karts = {}
	-- player karts have different physics
	for i = 1,playercount do
		karts[i] = {
			body = love.physics.newBody(world, points.karts[i][1], points.karts[i][2], "dynamic"),
			shape = love.physics.newRectangleShape(5, 5),
			traction = 1.0,
			jump = 0.0,
			bump = false,
			drifting = 0,
			skidding = 0,
			speed = {x=0, y=0, absolute=0, lastabsolute=0},
			terrain = "asphault",
			rumble = 0,
			wet = 0,
			gas = 0,
			brake = 0,
			steer = 0,
			sfxq = {},
		}
		karts[i].fixture = love.physics.newFixture(karts[i].body, karts[i].shape)
		karts[i].fixture:setRestitution(0.25)
		--karts[i].body:setAngle(track.start[3])
		karts[i].body:setAngle(-pi*.5)
		karts[i].x, karts[i].y = karts[i].body:getPosition()
		karts[i].lastx, karts[i].lasty = karts[i].body:getPosition()

		-- place raindrops
		if track.rain then
			raindrops[i] = {}
			local r = raindrops[i]
			for x = -10,10,0.5 do
				for y = -10,10,0.5 do
					r[#r+1] = {
						x = karts[i].x + x * 20,
						y = karts[i].y + y * 20,
						z = random(0,100)
					}
				end
			end
		end
	end

	-- CPU karts
	for i = playercount + 1,kartcount do
		karts[i] = {
			body = love.physics.newBody(world, points.karts[i][1], points.karts[i][2], "dynamic"),
			--ape = love.physics.newCircleShape(3),
			shape = love.physics.newRectangleShape(5, 5),
			traction = 1.0,
			jump = 0.0,
			bump = false,
			drifting = 0,
			skidding = 0,
			speed = {x=0, y=0, absolute=0, lastabsolute=0},
			terrain = "asphault",
			rumble = 0,
			wet = 0,
			gas = 0,
			brake = 0,
			steer = 0,
			target = {
				x = 0,
				y = 0,
			},
			sfxq = {},
		}
	--~ karts[i].fixture = love.physics.newFixture(karts[i].body, karts[i].shape,1.25) --- last value is density
	--~ karts[i].fixture:setRestitution(0)
	--~ --karts[i].body:setAngle(track.start[3])
	--~ karts[i].body:setAngle(-pi*.5)
	--~ karts[i].x, karts[i].y = karts[i].body:getPosition()
	--~ karts[i].lastx, karts[i].lasty = karts[i].body:getPosition()

		karts[i].fixture = love.physics.newFixture(karts[i].body, karts[i].shape)
		karts[i].fixture:setRestitution(0.25)
		karts[i].body:setAngle(-pi*.5)
		karts[i].x, karts[i].y = karts[i].body:getPosition()
		karts[i].lastx, karts[i].lasty = karts[i].body:getPosition()
	end

	if track.rain then sfx[1][1].rain:setVolume(0.75) end

	-- DEBUG performance test
	if false then
		for x = 1,32 do
			for y = 1,32 do
				coins[#coins + 1] = {x*8 - 16*8,-y*8}
			end
		end
	end

	collectgarbage()
end

function kartsystem.update(dt)
	timer = timer + dt
	world:update(dt) -- TODO is this better at the start or the end?

	if love.keyboard.isDown("1")  then camera.zoom = camera.zoom - dt*80 end
	if love.keyboard.isDown("2")  then camera.zoom = camera.zoom + dt*80 end
	if love.keyboard.isDown("3")  then camera.fov = camera.fov - dt end
	if love.keyboard.isDown("4")  then camera.fov = camera.fov + dt end
	if love.keyboard.isDown("5")  then camera.offset = camera.offset - dt end
	if love.keyboard.isDown("6")  then camera.offset = camera.offset + dt end

	for i = 1,#particles do
		local p = particles[i]
		if p then
			p.frame = p.frame - dt * p.speed
			if p.frame <= 0 then table.remove(particles,i) end
		end
	end

	for i = 1,#splashes do
		local s = splashes[i]
		if s then
			s.age = s.age + dt
			if s.age >= 1 then table.remove(splashes, i) end
		end
	end

	for p = 1,#karts do
		local kart = karts[p]
		--kart.wet = 0.3 -- DEBUG
		kart.lastx, kart.lasty = kart.x, kart.y
		kart.x,kart.y = kart.body:getPosition()
		kart.speed.lastabsolute = kart.speed.absolute
		kart.rumble = (kart.rumble + max(kart.speed.absolute/10, 7) * dt)%1
		kart.r = (kart.body:getAngle() + pi) % tau - pi
		kart.vecr = {x = cos(kart.r), y = sin(kart.r)}

		local sfx = sfx[1][p] -- TODO ditch this

		-- players only
		if p <= playercount then
			if isDown(p, "gas") then
				kart.gas = 1
				kart.brake = 0
			else
				kart.gas = 0
			end

			if isDown(p, "brake") then
				kart.brake = min(kart.brake + dt * 5, 1)
				kart.gas = 0
			else
				kart.brake = 0
			end

			if isDown(p, "left") and not isDown(p, "right") then
				kart.steer = -1
			elseif isDown(p, "right") and not isDown(p, "left") then
				kart.steer = 1
			else
				kart.steer = 0
			end
			if controls.controllers[p] then
				if controls.controllers[p]:getGamepadAxis("leftx") ~= 0 then
					kart.steer = controls.controllers[p]:getGamepadAxis("leftx")
				end
				if controls.controllers[p]:getGamepadAxis("triggerright") > 0 then
					kart.gas = controls.controllers[p]:getGamepadAxis("triggerright")
				end
				if controls.controllers[p]:getGamepadAxis("triggerleft") > 0 then
					kart.brake = controls.controllers[p]:getGamepadAxis("triggerleft")
				end
			end
		end

		-- collect coins
		for i = 1,#coins do
			if dist(kart.x,kart.y,coins[i][1],coins[i][2]) < 8 then -- TODO make macro
				table.remove(coins, i)
				kart.sfxq.coin = 1
				break
			end
		end

		if kart.x >= 0 and kart.x < track.w and kart.y >= 0 and kart.y < track.h then
			local r,g,b = track.terrain:getPixel(kart.x, kart.y)
			kart.terrain = tracktype[ceil(r*2)*9 + ceil(g*2)*3 + ceil(b*2)*1]
		end

		-- jump pads
		if kart.terrain == "bumper" and kart.jump == 0 then
			kart.jump = 1
			kart.sfxq.jump = 1
			kart.bump = true
			kart.fixture:setFilterData(1,0,0) -- disable physics

			x,y = vector.rotate(max( (200 - kart.speed.x)*5, 1),0,kart.r )
			kart.body:applyForce(x, y)
		end

		-- normal jump
		if kart.jump == 1 then
			vibrate(p, 1.0, 1.0, 0.05)
			--TODO is this even good?
			-- give boost to drift hops
			local x,y = cos(kart.r + pi * .5), sin(kart.r + pi * .5)
			local s = 0
			if kart.drifting == 1 then
				--s = min(-200 + kart.speed.absolute, 0)
				s = -100
			elseif kart.drifting == -1 then
				--s = max( 200 - kart.speed.absolute, 0)
				s = 100
			end
			s = s * kart.gas
			kart.body:applyForce(x*s,y*s)
			-- jump in place stronger when going slow
			if kart.speed.x < 75 and kart.speed.x >= 0 and kart.speed.absolute < 75 then
				local x,y = vector.rotate((75 - kart.speed.x)*10 * kart.gas,0,kart.r )-- - pi * 0.5 * kart.drifting)
				kart.body:applyForce(x, y)
			end
			kart.body:applyAngularImpulse(kart.steer*.25)
		end

		if kart.jump > 0 then
			kart.traction = 0.0
		elseif kart.terrain == "dirt" then
			kart.traction = 0.25
		else
			kart.traction = 1
		end
		if track.rain then kart.wet = max(kart.wet, 0.25) end

		-- water
		if kart.jump == 0 and kart.terrain == "water" then
			kart.wet = 1
		elseif kart.wet > 0 then
			kart.wet = max(kart.wet - dt, 0)
			--kart.traction = kart.traction * (1 - pow(kart.wet*(kart.speed.absolute/200), 0.5)*.75)
			kart.traction = kart.traction * (1 - pow(kart.wet*(kart.speed.absolute/200), 0.25)*.75)
			if kart.jump == 0 then --and kart.wet > 0.25 then
				kart.body:applyTorque( clamp(kart.body:getAngularVelocity() * kart.wet * 0.25, -1, 1) * (kart.speed.absolute/10) )
			end
		end

		-- better traction for AI
		--~ if p > playercount then
			--~ kart.traction = min(kart.traction * 1.5, 1)
		--~ end



		if kart.jump > 0 then
			-- jump from bumper/jump pad lasts longer
			if kart.bump then
				kart.jump = kart.jump - dt*3
			else
				kart.jump = kart.jump - dt*4
			end

			-- jump landing
			if kart.jump <= 0 then
				kart.sfxq.land = 1
				vibrate(p, 0.0, 1.0, 0.05)
				kart.jump = 0
				kart.bump = false
				kart.fixture:setFilterData(1,0xffff,0)
				--~ if kart.drifting == 0 then
					--~ kart.body:setAngularVelocity(clamp(kart.body:getAngularVelocity()*2 * clamp(kart.speed.absolute/200, 0, 1), -2, 2))
				--~ end

				if abs(kart.body:getAngularVelocity()) * (kart.speed.absolute / 100) > 2 then
					if kart.body:getAngularVelocity() > 0 then
						kart.drifting = 1
					else
						kart.drifting = -1
					end
				end
			end
		end

		if kart.jump > 0 then
			kart.body:setLinearDamping(0)
		elseif kart.terrain == "dirt" then
			kart.body:setLinearDamping(2)
		elseif kart.terrain == "water" then
			kart.body:setLinearDamping(0.125)
		else
			kart.body:setLinearDamping(.25)
		end
		kart.body:setAngularDamping(kart.traction * 10)

		local x1,y1 = kart.body:getLinearVelocity()
		kart.speed.x,kart.speed.y = vector.rotate(x1, y1, -kart.r)
		kart.speed.absolute = vector.length(x1,y1)

		kart.skidding = clamp(clamp(abs(kart.speed.y)/abs(kart.speed.x) - 0.25, 0, 1), 0, 1) -- clamp(1 - kart.speed.absolute/200,0,1), 0, 1)
		local d = abs(kart.speed.lastabsolute - kart.speed.absolute)
		if d > 2 then
			kart.skidding = clamp(kart.skidding + d*0.1, 0, 1)
		end

		-- glue to the ground, lessens when drifting
		local x2,y2 = cos(kart.r - pi * .5), sin(kart.r - pi * .5)
		if kart.drifting == 0 then
			kart.body:applyForce(
				x2*kart.speed.y*kart.traction*.25,
				y2*kart.speed.y*kart.traction*.25
			)
		else
			kart.body:applyForce(
				x2*kart.speed.y*kart.traction*0.1,
				y2*kart.speed.y*kart.traction*0.1
			)
		end

		-- disable drifting
		--if kart.drifting == 1 and not isDown(p, "right") then kart.drifting = 0 end
		--if kart.drifting == -1 and not isDown(p, "left") then kart.drifting = 0 end
		if kart.speed.absolute < 75 or kart.speed.x <= 0 then kart.drifting = 0 end
		if kart.drifting > 0 and kart.body:getAngularVelocity() < 0 then kart.drifting = 0 end
		if kart.drifting < 0 and kart.body:getAngularVelocity() > 0 then kart.drifting = 0 end
		if kart.skidding < 0.05 then kart.drifting = 0 end
		--if kart.wet > 0.1 then kart.drifting = 0 end


		if kart.jump == 0 then
			if kart.gas > 0 then
				local s = max(kart.gas * 30 - max(kart.speed.x, 0)*.125, 0)
				-- apply a little forward force when drifting
				if kart.drifting ~= 0 then
					s = s * 1.25
				end
				kart.body:applyForce(kart.vecr.x*s, kart.vecr.y*s)
			elseif kart.brake > 0 then
				if kart.speed.absolute > 10 then
					local x,y = 0,0
					local s = kart.traction
					x,y = vector.rotate(kart.brake * -kart.speed.x,0,kart.r)
					kart.body:applyForce(x*s, y*s)
				elseif kart.speed.absolute > 0 then
					kart.body:setLinearVelocity(0,0)
				end

				kart.body:setLinearDamping(kart.body:getLinearDamping() * 2)
				kart.body:setAngularDamping(kart.body:getAngularDamping() * 2)
			end

			-- steering
			if abs(kart.speed.x) > 5 then
				local s = (2.75 - kart.speed.absolute/125) * kart.traction * 3
				-- steer faster when drifting
				if kart.drifting > 0 and kart.steer > 0
				or kart.drifting < 0 and kart.steer < 0
				then s = s * 2 end
				if kart.speed.x < 0 then s = -s end
				kart.body:applyTorque(s * kart.steer)
			end

		end

		if kart.speed.lastabsolute - kart.speed.absolute > 10 then
			local hard = clamp((kart.speed.lastabsolute - kart.speed.absolute)/100, 0, 1)
			kart.sfxq.collision = hard
			if p <= playercount then vibrate(p, hard, hard, 0.1) end
		end
	end

	-- player ------------------------------------------------------------------
	for p = 1,playercount do
		local kart = karts[p]

		if track.rain then
			for i = 1,#raindrops[p] do
				local r = raindrops[p][i]
				if r then
					r.z = r.z - dt * 100
					--snow
					--~ r.x = r.x + random(-25,25)*dt
					--~ r.y = r.y + random(-25,25)*dt
					if r.z <= 0 then
						splashes[#splashes+1] = {
							x = r.x,
							y = r.y,
							age = 0.7
						}
						--snow
						--~ canvas.tiremarks:renderTo(function()
							--~ lg.points(r.x,r.y)
						--~ end)
						                                   -- place toward where the kart is headed
						r.x = random(-200, 200) + kart.x + (kart.x - kart.lastx)*50
						r.y = random(-200, 200) + kart.y + (kart.y - kart.lasty)*50
						r.lx = nil
						r.ly = nil
						r.z = r.z + 100
					end
				end
			end
		end
	end

	-- end player --------------------------------------------------------------

	-- AI ----------------------------------------------------------------------
	for i = playercount + 1,#karts do
		local kart = karts[i]
		-- find if we're in a box
		local ai = track.ai
		for i = 1,#ai do
			local a = ai[i]
			if kart.x > a[1] and kart.x < a[3] and kart.y > a[2] and kart.y < a[4] then
				kart.target = {x = a[5], y = a[6]}
			end
		end
		local ass = atan2(kart.target.y - kart.y, kart.target.x - kart.x) - kart.r - pi * .5
		-- hop to turn around
		if abs(ass) > 0.5 and kart.jump == 0 and kart.speed.absolute < 10 then kart.jump = 1 end
		local target = 0
		if track.rain then -- steer slower,  correct for angular changes more, throttle less
			target = atan2(cos(ass), sin(ass))*1.5 - kart.body:getAngularVelocity()*0.5
			kart.gas = clamp(1 - abs(kart.steer)*.025, 0.1, 1)
		else
			target = atan2(cos(ass), sin(ass))*1.5 - kart.body:getAngularVelocity()*0.25
			kart.gas = clamp(1 - abs(kart.steer)*.025, 0.1, 1)
		end
		kart.steer = target
	end
	-- end AI ------------------------------------------------------------------

	-- SFX ---------------------------------------------------------------------
	for p = 1,playercount do
		for k = 1,kartcount do
			local kart = karts[k]
			local sfx = sfx[p][k]
			local x,y = vector.rotate(kart.x - karts[p].x, kart.y - karts[p].y, -karts[p].r - 0.5*pi)
			local x2,y2 = x/32,y/32
			local d = doppler[p][k]
			local dx,dy = (x-d.x)*20, (y-d.y)*20

			-- most SFX, such as jumping, coins etc
			for f,k in pairs(kart.sfxq) do
				sfx[f]:setVolume(kart.sfxq[f])
				sfx[f]:setPosition(x2, y2, 0)
				sfx[f]:setVelocity(dx, dy, 0)
				sfx[f]:stop()
				sfx[f]:play()
			end

			sfx.screech:setVolume(clamp(kart.skidding * (1 - kart.wet^0.1), 0, 0.5))
			sfx.screech:setPitch(clamp(kart.speed.absolute/200, 0.5, 1))
			sfx.screech:setPosition(x2, y2, 0)
			if kart.jump > 0 or kart.terrain ~= "asphault" then sfx.screech:setVolume(0) end
			sfx.screech:setVelocity(dx, dy, 0)

			sfx.engine:setVolume(clamp((kart.speed.x/400) * (0.5 + kart.gas * 0.5), 0.1, 0.5))
			sfx.engine:setPitch(clamp(kart.speed.x/50 - 1, 0.5, 4.0))
			if kart.speed.x < 50 then
				sfx.engine:setPitch(clamp(1 - kart.speed.x/50, 0.5, 0.75))
			end
			sfx.engine:setPosition(x2, y2, 0)
			sfx.engine:setVelocity(dx, dy, 0)

			if kart.jump == 0 and kart.terrain == "rumble" then
				sfx.rumble:setVolume(clamp(kart.speed.absolute/500 - 0.1, 0, 1.0))
				sfx.rumble:setPitch(clamp(kart.speed.absolute/200 + 0.5, 0.5, 4.0))
			else
				sfx.rumble:setVolume(0)
			end
			sfx.rumble:setPosition(x2, y2, 0)
			sfx.rumble:setVelocity(dx, dy, 0)

			if kart.jump == 0 and kart.terrain == "water" then
				sfx.water:setVolume(clamp(kart.speed.absolute/500 - 0.1, 0, 1.0))
				sfx.water:setPitch(clamp(kart.speed.absolute/200 + 0.5, 0.5, 4.0))
			else
				sfx.water:setVolume(0)
			end
			sfx.water:setPosition(x2, y2, 0)
			sfx.water:setVelocity(dx, dy, 0)


			if kart.jump == 0 and kart.terrain == "dirt" then
				sfx.dirt:setVolume(clamp(kart.speed.absolute/500, 0, 1.0))
				sfx.dirt:setPitch(clamp(kart.speed.absolute/500 + 0.5, 0.5, 4.0))
			else
				sfx.dirt:setVolume(0)
			end
			sfx.dirt:setPosition(x2, y2, 0)
			sfx.dirt:setVelocity(dx, dy, 0)


			doppler[p][k] = {x = x, y = y}
		end

	-- Rumble ------------------------------------------------------------------
		local kart = karts[p]
		if kart.jump == 0 then
			if kart.terrain == "dirt" then
				vibrate(p, 0.0, clamp(kart.speed.absolute/200, 0, 0.5))
			elseif kart.terrain == "rumble" then
				vibrate(p, clamp(kart.speed.absolute/200, 0, 0.5), 0.0)
			else
				vibrate(p, 0.0, 0.0)
			end
		else
			vibrate(p, 0.0, 0.0)
		end
	end

	-- clear SFX queue
	for i = 1,kartcount do karts[i].sfxq = {} end
end


-- it's faster to draw background twice than switch canvases
local function bgdraw(reflection, kart)
	lg.push()
	lg.translate(0, rendersize.bg)
	if reflection then lg.scale(1,-1) end
	local fov = (camera.fov/camera.offset)
	for i = 1,#track.bg do
		local b = track.bg[i]
		for j = 0,ceil(rendersize.w/ (b.w/camera.bgscale) ) + 1 do
			lg.draw(
				b.tex,
				((-kart.r/tau) * b.w + b.w * j)*camera.bgscale,
				0,
				0, camera.bgscale, nil, b.w, b.h
			)
		end
	end
	lg.pop()
end

--~ function mix(from, to, val)
	--~ return {
		--~ from[1] + (to[1] - from[1]) * val,
		--~ from[2] + (to[2] - from[2]) * val,
		--~ from[3] + (to[3] - from[3]) * val,
	--~ }
--~ end

function kartsystem.draw()

	--draw skidmarks
	lg.setCanvas(canvas.tiremarks)
	for i = 1,kartcount do
		local kart = karts[i]

		local tirelx, tirely = vector.rotate(0,  2.5, kart.r)
		local tirerx, tirery = vector.rotate(0, -2.5, kart.r)
		tirelx = tirelx + kart.x
		tirely = tirely + kart.y
		tirerx = tirerx + kart.x
		tirery = tirery + kart.y

		-- tire smoke
		if kart.wet == 0 then
			if kart.skidding > 0 and kart.jump == 0 and kart.terrain == "asphault" then
				lg.setColor(0,0,0, kart.skidding)
				lg.points(tirelx,tirely);lg.points(tirerx,tirery)
				particles[#particles+1] = {
					x = tirelx + random(-1,1),
					y = tirely + random(-1,1),
					frame = 4,
					speed = 20,
					scale = (kart.speed.absolute/800) * (kart.skidding) + 0.125,
					name = "smoke"
				}
				particles[#particles+1] = {
					x = tirerx + random(-1,1),
					y = tirery + random(-1,1),
					frame = 4,
					speed = 20,
					scale = (kart.speed.absolute/800) * (kart.skidding) + 0.125,
					name = "smoke"
				}
			end

		-- water skidding
		elseif kart.jump == 0 and kart.speed.absolute > 10 and kart.wet > 0 then
			local s = (kart.speed.absolute/800 + 0.125)*max(kart.wet, 0.5) * max(kart.skidding*4, 1)
			particles[#particles+1] = {
				x = tirelx + random(-1,1)*.5,
				y = tirely + random(-1,1)*.5,
				frame = 3,
				speed = 20,
				scale = s,
				name = "splash"
			}
			particles[#particles+1] = {
				x = tirerx + random(-1,1)*.5,
				y = tirery + random(-1,1)*.5,
				frame = 3,
				speed = 20,
				scale = s,
				name = "splash"
			}
		end

		-- dirt particles
		if kart.terrain == "dirt" and kart.jump == 0 and kart.speed.absolute > 10 then
			lg.setColor(0.4,0.3,0.1, 0.5)
			lg.points(tirelx,tirely);lg.points(tirerx,tirery)

			particles[#particles+1] = {
				x = tirelx + kart.vecr.x*random(-2,2),
				y = tirely + kart.vecr.y*random(-2,2),
				frame = 4,
				speed = 20,
				scale = kart.speed.absolute/800 + 0.125,
				name = "dirt"
			}
			particles[#particles+1] = {
				x = tirerx + kart.vecr.x*random(-2,2),
				y = tirery + kart.vecr.y*random(-2,2),
				frame = 4,
				speed = 20,
				scale = kart.speed.absolute/800 + 0.125,
				name = "dirt"
			}
		end

		if kart.terrain == "water" and kart.jump == 0 and kart.speed.absolute > 10 then
			splashes[#splashes+1] = {
				x = tirelx,
				y = tirely,
				age = 1 - kart.speed.absolute/200
			}
			splashes[#splashes+1] = {
				x = tirerx,
				y = tirery,
				age = 1 - kart.speed.absolute/200
			}
		end
	end

	--shadows
	lg.setCanvas(canvas.shadow)
	lg.clear()
	lg.setColor(0.25, 0.25, 0.25, 1.0)
	for i = 1,kartcount do
		lg.circle("fill", karts[i].x, karts[i].y, 3.5*(0.75 + abs(karts[i].jump-0.5)*.5), 8)
	end
	lg.setColor(1, 1, 1, 1)

	lg.setCanvas(canvas.track)
	lg.clear()
	lg.draw(track.tex.main)

	lg.draw(canvas.tiremarks)

	lg.setBlendMode("subtract")
	lg.draw(canvas.shadow)
	lg.setBlendMode("alpha")

	-- water splashes
	for i = 1,#splashes do
		local s = splashes[i]
		lg.draw(tex.ripple, s.x, s.y, 0, (1 - s.age)*.5, nil, 8, 8)
	end


	for p = 1,playercount do

		local sprite = {}
		local player = players[p]
		local kart = karts[p]



		local function drawSprites(mirror)
			lg.setShader(shader.fog)
			table.sort(sprite,function(a,b) return a.dist>b.dist end)
			for i = 1,#sprite do
				local draw = sprite[i]
				local x,y = vector.rotate((kart.x-draw.x)/camera.zoom, (kart.y-draw.y)/camera.zoom, -kart.r + 0.5 * pi)
				y = y * camera.fov
				local dist = 1 - y
				local size = ((1/dist)/camera.zoom*camera.offset)*rendersize.w
				if size > 0 then
					local scale = draw.size or 1
					local height = (draw.jump or 0) * mirror
					local x = (x/dist)*camera.offset*rendersize.w + rendersize.w/2
					local y = ( (y+camera.offset-1)/dist )*rendersize.track + rendersize.track + height * size + rendersize.bg
					--lg.setColor(mix({0,0,0}, track.fogcol, (1/dist) * track.fogcol[4]))
					--lg.setColor(mix({1,1,1}, track.fogcol, min(dist/16 - 0.1, 1)))
					shader.fog:send("dist", min((dist - 1)/8, 1) * track.fogcol[4])
					if draw.quad then
						lg.draw(draw.tex, draw.quad, x, y, 0,
							scale*size, mirror*scale*size,
							draw.offset, draw.offset*2
						)
					else
						lg.draw(draw.tex, x, y, 0,
							scale*size, mirror*scale*size,
							draw.offset, draw.offset*2
						)
					end
				end
			end
			lg.setShader()
			--lg.setColor(1,1,1)
		end


		shader.track:send("x", kart.x/track.w)
		shader.track:send("y", kart.y/track.h)
		shader.track:send("zoom", camera.zoom)
		shader.track:send("offset", camera.offset)
		shader.track:send("fov", camera.zoom/camera.fov)
		shader.track:send("rotation", {sin(-kart.r), cos(-kart.r), cos(-kart.r), -sin(-kart.r)})

		shader.far:send("x", kart.x/track.w)
		shader.far:send("y", kart.y/track.h)
		shader.far:send("zoom", camera.zoom)
		shader.far:send("fov", camera.zoom/camera.fov)
		shader.far:send("offset", camera.offset)
		shader.far:send("rotation", {sin(-kart.r), cos(-kart.r), cos(-kart.r), -sin(-kart.r)})


		-- sprites that get reflected-------------------------------------------

		-- players
		for k = 1,kartcount do
			local kart2 = karts[k]
			local q = 0
			if k == p then -- main kart on screen
				q = math.floor( clamp(kart2.speed.y/max(kart2.speed.x, 0.01), -2, 2) * -2 + 0.5)%22
			else
				q = clamp(math.floor( ( ( (-kart.r + kart2.r)/tau + 1/44)%1)*22 ), 0, 21)
			end
			sprite[#sprite+1] = {
				dist = (kart2.x - kart.x) * cos(kart.r) + (kart2.y - kart.y) * sin(kart.r),
				tex = players[k].tex,
				quad = kartquad[q],
				x = kart2.x,
				y = kart2.y,
				jump = cos(kart2.jump*pi + pi*.5) * 4 + floor(kart2.rumble*2)*.25,
				offset = 16,
				size = 1/4
			}
		end

		-- hazards
		for i = 1,#hazards do
			local p = hazards[i]
			sprite[#sprite+1] = {
				dist = (p[1] - kart.x) * cos(kart.r) + (p[2] - kart.y) * sin(kart.r),
				x = p[1],
				y = p[2],
				tex = tex.hazard[p[3]],
				offset = tex.hazard[p[3]]:getHeight() * .5, -- TODO this is crap
				size = p[4]
			}
		end

		-- coins
		for i = 1,#coins do
			local p = coins[i]
			sprite[#sprite+1] = {
				dist = (p[1] - kart.x) * cos(kart.r) + (p[2] - kart.y) * sin(kart.r),
				x = p[1],
				y = p[2],
				tex = tex.coin,
				offset = 8,
				size = 1/4
			}
		end



		-- DEBUG AI targets
		--~ for i = playercount + 1, kartcount do
			--~ local k = karts[i]
			--~ sprite[#sprite+1] = {
				--~ dist = (k.target.x - kart.x) * cos(kart.r) + (k.target.y - kart.y) * sin(kart.r),
				--~ x = k.target.x,
				--~ y = k.target.y,
				--~ tex = tex.coin,
				--~ offset = 8
			--~ }
		--~ end
		-- /sprites ------------------------------------------------------------

		-- draw track shadow over player shadows,
		-- only for bottom layer if it exists
		if track.tex.shadow then
			lg.setCanvas(canvas.shadow)
			lg.draw(track.tex.shadow)
		end

		-- sky
		lg.setCanvas(canvas.screen[p])
		if not track.reflection then lg.clear() end
		--~ lg.setColor(0xf8/255,0xe8/255,0x90/255)
		--~ lg.rectangle("fill", 0, 0, rendersize.w, rendersize.h)
		--~ lg.setColor(1,1,1)
		if track.tex.gradient then
			lg.draw(track.tex.gradient, 0, 0, 0, (rendersize.w/256), (rendersize.bg/256))
		end
		if track.reflection then
			-- TODO not fixed 256x256
			lg.draw(track.tex.reflection, 0, rendersize.bg, 0, (rendersize.w/256), (rendersize.track/256))
		end

		-- clouds
		if track.tex.sky then
			shader.clouds:send("x", -kart.r/tau + (timer%128)/128)
			lg.setShader(shader.clouds)
			lg.draw(canvas.cloudsdummy)
			if track.reflection then
				lg.draw(canvas.cloudsdummy, 0, rendersize.h, 0, 1, -1)
			end
			lg.setShader()
		end

		if track.reflection then
			-- background in reflection
			bgdraw(true, kart)
			-- reflected sprites
			-- all sprites added after this point are not reflected
			drawSprites(-1)
		end

		--draw track in psudo-3D


		-- lava
		local mul = 1 + (1/camera.zoom) * 16 -- TODO make this not hard coded 16 pixels
		if track.tex.lava then
			lg.setShader(shader.far)

			shader.far:send("map", track.tex.lava)
			shader.far:send("nocenter", false)
			shader.far:send("scale", 128)
			shader.far:send("x", ((kart.x + (timer%2)*4)/track.w))
			shader.far:send("zoom", camera.zoom * mul)
			shader.far:send("offset", camera.offset * mul)

			lg.draw(canvas.trackdummy, 0, rendersize.bg)
		end


		-- bottom shadow
		if track.tex.shadow then
			lg.setShader(shader.track)
			shader.track:send("zoom", camera.zoom * mul)
			shader.track:send("offset", camera.offset * mul)
			--shader:send("fov", (camera.zoom/camera.fov) * 1.5)

			shader.track:send("map", canvas.shadow)
			lg.setBlendMode("subtract")
			lg.draw(canvas.trackdummy, 0, rendersize.bg)
			lg.setBlendMode("alpha")
		end

		-- distant repeating texture
		lg.setShader(shader.far)
			shader.far:send("map", track.tex.far)
			shader.far:send("nocenter", true)
			shader.far:send("scale", 128)
			shader.far:send("x", kart.x/track.w)

			shader.track:send("zoom", camera.zoom)
			shader.track:send("offset", camera.offset)

			lg.draw(canvas.trackdummy, 0, rendersize.bg)

			-- track
		lg.setShader(shader.track)
			shader.track:send("map", canvas.track)
			shader.track:send("zoom", camera.zoom)
			shader.track:send("offset", camera.offset)
			lg.draw(canvas.trackdummy, 0, rendersize.bg)

		lg.setShader()


		--background, non-reflected
		bgdraw(false, kart)

		-- sprites that don't get reflected ------------------------------------
		-- particles
		for i = 1,#particles do
			local p = particles[i]
			local p2 =  particle[p.name]
			sprite[#sprite+1] = {
				dist = (p.x - kart.x) * cos(kart.r) + (p.y - kart.y) * sin(kart.r),
				x = p.x,
				y = p.y,
				tex = p2.tex,
				quad =  p2.quads[ceil(p.frame)],
				size = (p.frame / p.speed*2) * p.scale,
				offset = p2.offset
			}
		end
		-- /sprites ------------------------------------------------------------

		drawSprites(1)

		-- rain
		if track.rain then
			-- rain gets more transparent the faster you spin, to make up for the longer raindrops
			-- also mitigates things going crazy on the frame you hit a wall
			lg.setColor(0.75,0.9,1.0, max( 0.5 - abs(kart.body:getAngularVelocity()) * 0.125 , 0))
			for i = 1,#raindrops[p] do
				local draw = raindrops[p][i]
				local x,y = vector.rotate((kart.x-draw.x)/camera.zoom, (kart.y-draw.y)/camera.zoom, -kart.r + 0.5*pi)
				y = y * camera.fov
				local dist = 1 - y
				local size = ((1/dist)/camera.zoom*camera.offset)*rendersize.w
				if size > 0 then
					local x = (x/dist)*camera.offset*rendersize.w + rendersize.w/2
					local y = ( (y+camera.offset-1)/dist )*rendersize.track + rendersize.track - draw.z * size + rendersize.bg
					if draw.lx then lg.line(x, y, draw.lx, draw.ly) end
					--if draw.lx then lg.rectangle("fill", x, y, 1, 10) end
					--~ lg.points(x,y) -- snow
					draw.lx, draw.ly = x, y
				end
			end
			lg.setColor(1,1,1)
		end

		lg.print(floor(kart.speed.absolute))
		--lg.print(serial.pack(camera))
	end


	lg.setCanvas()
	return canvas.screen
end

function kartsystem.draw_overlay()
	if love.keyboard.isDown("l") then
		lg.center(track.w, track.h)
		lg.draw(canvas.track)
		for i = 1,kartcount do
			local kart = karts[i]
			--love.graphics.polygon("fill", kart.body:getWorldPoints(kart.shape:getPoints()))
			--lg.line(kart.x, kart.y, kart.x+kart.vecr.x*16, kart.y+kart.vecr.y*16)
			lg.draw(
				players[i].tex,
				kartquad[clamp(math.floor( ( ( (kart.r)/tau + .25)%1)*22 ), 0, 21)],
				kart.x, kart.y,
				0,
				1, 1,
				16, 32
			)
		end
	end
end

function kartsystem.press(key, player)
	local kart = karts[player]
	if key == "jump" and kart.jump == 0 then
		kart.jump = 1
		kart.sfxq.jump = 1
	end
end

return kartsystem