lg = love.graphics
lg.setDefaultFilter("nearest", "nearest")
love.graphics.setLineStyle("rough")
-- why is this hack necessary on android, I have high DPI turned off...
if love.system.getOS() == "Android" then
	local s = love.graphics.newCanvas
	lg.newCanvas = function(w,h) return s(w,h,{dpiscale = 1}) end
end

require "maths"
menu = require "menu"

state = "menu"

frametime = 0
redraw = true
fixedfps = false

playercount = 1

canvas_screen = {}

function lg.center(w,h)
	local s = min(screen.w/w, screen.h/h)
	local x = floor((screen.w - w*s)*0.5)
	local y = floor((screen.h - h*s)*0.5)
	lg.translate(x,y)
	lg.scale(s)
end

function vibrate(player, str1, str2, dur)
	if controls.controllers[player] then
		if controls.vibtime[player] == 0 then
			controls.controllers[player]:setVibration(str1, str2, dur)
		end
		if dur then controls.vibtime[player] = dur end
	end
end

function gameinit(options)
	kartsystem.init(options)
end

function love.load()

	controls = {
		mapping = {
			keyboard_inv = {}, -- auto generated
			keyboard = {
				{
					gas = {"up"},
					brake = {"down"},
					left = {"left"},
					right = {"right"},
					jump = {"space"},
				},
				{
					gas = {"w", ","},
					brake = {"s", "o"},
					left = {"a"},
					right = {"d", "e"},
					jump = {"lshift"},
				}
			},
			controller_inv = {}, -- auto generated
			controller = {
				{
					gas = {"a"},
					brake = {"x"},
					left = {"dpleft",},
					right = {"dpright",},
					jump = {"rightshoulder", "b"},
				},
				{
					gas = {"a"},
					brake = {"x"},
					left = {"dpleft",},
					right = {"dpright",},
					jump = {"rightshoulder", "b"},
				}
			},
		},
		 -- WARNING this table can have holes after controllers are assigned to players
		controllers = love.joystick.getJoysticks(),
		vibtime = {}, -- auto generated
		assign = {
			"Sony Computer Entertainment Wireless Controller",
		},
	}

	for i,k in pairs(controls.controllers) do
		controls.vibtime[i] = 0
	end

	-- TODO untested
	-- swap controllers around such that controller names match player
	local c = {}
	for i = 1,#controls.controllers do for j = 1,#controls.assign do
		if controls.controllers[i]:getName() == controls.assign[j] and j ~= i then
			controls.controllers[i], controls.controllers[j] =
			controls.controllers[j], controls.controllers[i]
		end
	end end

	-- invert control tables, so gas = {x, y} becomes x = gas, y = gas
	-- TODO run this when controls are changed
	for player,c in ipairs(controls.mapping.keyboard) do
		controls.mapping.keyboard_inv[player] = {}
		for i, con in pairs(c) do
			for k,l in pairs(con) do
				controls.mapping.keyboard_inv[player][l] = i
			end
		end
	end
	for player,c in ipairs(controls.mapping.controller) do
		controls.mapping.controller_inv[player] = {}
		for i, con in pairs(c) do
			for k,l in pairs(con) do
				controls.mapping.controller_inv[player][l] = i
			end
		end
	end


	kartsystem = require "kartsystem"
	love.resize()

	shader = {}
	for i,j in pairs(love.filesystem.getDirectoryItems("shaders")) do
		shader[j:sub(1,-6)] = lg.newShader("shaders/" .. j)
	end

	menu.init()


	-- skip menu
	if false then
		state = "game"
		gameinit({
			player1kart = "mario",
			player2kart = "luigi",
			track = "luigi",
			ai = 0,
			players = 1,
			splitscreen = "tall",
			resolution = {
				w = 400, h = 225, bg = 80,
				offset = 0.7, fov = 1.0,
				kartscale = 1.0, bgscale = 1.0
			},
		})
	end

end

function gameupdate(dt)
	for i,k in pairs(controls.vibtime) do
		controls.vibtime[i] = max(k - dt, 0)
	end

	kartsystem.update(dt)
end

function love.update(dt)
	if state == "game" then
		gameupdate(dt)
		redraw = true
		-- TODO, this is disabled for now
		--~ if not fixedfps then
			--~ frametime = frametime - dt
			--~ while frametime <= 0 do
				--~ frametime = frametime + 1/60
				--~ gameupdate(1/60)
				--~ redraw = true
			--~ end

			--~ --if dt < 1/60 then love.timer.sleep(1/120) end
		--~ else
			--~ redraw = true
			--~ --gameupdate(1/60)
			--~ gameupdate(dt)
		--~ end
	elseif state == "menu" then
		menu.update(dt)
	end
end

function love.draw()
	lg.push()
	if state == "game" then
		if redraw then
			redraw = false
			canvas_screen = kartsystem.draw()
		end

		for p = 1,playercount do
			local w,h = canvas_screen[p]:getWidth(),canvas_screen[p]:getHeight()
			lg.push();
			if playercount == 2 then
				if w > h then -- TODO actually grab camera.splitscreen from kartsystem.lua
					lg.center(w,h * 2)
					lg.draw(canvas_screen[p], 0, (p-1)*h)
				else
					lg.center(w*2,h)
					lg.draw(canvas_screen[p], (p-1)*w, 0)
				end
			else
				lg.center(w,h)
				lg.draw(canvas_screen[p])
			end
			lg.pop()
		end

		kartsystem.draw_overlay()

	elseif state == "menu" then
		menu.draw()
	end
	lg.pop()

	-- DEBUG
	if false then
		lg.setColor(0,0,0,0.5)
		lg.rectangle("fill", 0, 0, 1024, 1024)
		lg.setColor(1,1,1)
		lg.print("FPS: " .. love.timer.getFPS(), 0, 0, 0, 8)
		lg.print("RAM: " .. floor(collectgarbage("count")/1024 + 0.5) .. "MB", 0, 96, 0, 8)
		lg.print(serial.pack(love.graphics.getStats()), 0, 192, 0, 4)
	end
end

function love.keypressed(key)
	if key =="escape" then love.event.quit() end
	if state == "game" then
		for i = 1, min(#controls.mapping.keyboard_inv, playercount) do
			kartsystem.press(controls.mapping.keyboard_inv[i][key], i)
		end
		-- DEBUG
		if key == "b" then
			--love.window.setMode(lg.getWidth(), lg.getHeight(), {vsync=false})
		end
	elseif state == "menu" then
		local s = menu.press(key)
		if s ~= nil then
			state = "game"
			gameinit(s)
		end
	end
end

function love.gamepadpressed(joystick, button)
	if state == "game" then
		for i = 1, min(#controls.mapping.controller_inv, playercount) do
			kartsystem.press(controls.mapping.controller_inv[i][button], i)
		end
	end
end

function isDown(player, key)
	if player <= playercount then
		local bool = false
		bool = love.keyboard.isDown(unpack(controls.mapping.keyboard[player][key]))
		if controls.controllers[player] then
			bool = controls.controllers[player]:isGamepadDown(unpack(controls.mapping.controller[player][key])) or bool
		end
		return bool
	end
end

function love.resize(w, h)
	w = w or lg.getWidth()
	h = h or lg.getHeight()
	screen = {
		w = w,
		h = h,
	}
end